% Description: Create a summary of the observations based on the
%  direction of communciation.
% Author(s) : Xinming Ou, Sakthiyuvaraja Sakthivelmurugan, Loai Zomlot
% Copyright (C) 2012, Argus Cybersecurity Lab, Kansas State University

% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.

% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.

% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.


:- import db_query/4, db_prepare/3, db_prepare_execute/3, db_statement_close/1 from db_interface.
:- import concat_atom/2 from string.


%Find summary of all alerts.
summarize:-
	getStartPoint(StartPoint),
	counter_init_seed(sk,StartPoint),
	findall(Mode1, summary(int(_Pred2(_H1, _H2, Mode1,_,_))), _L1), 
	findall(Mode2, summary(int(_Pred1(_Host, Mode2,_,_))), _L2), %compromised predicate
	insertSummaryInDB,
	insertMappingInDB.

summary(int(Pred(H1, H2, Mode, skolem(N, SnortRuleID), TimeRange))):-
	setof(oid(SID,CID), Time^deriveIntPred(int(Pred(H1, H2), Mode, Time), oid(SID,CID), SnortRuleID), ListOfOID),
	counter_next(sk, N),
	getTimeRangeForSummary(ListOfOID, TimeRange),
	assert(int(Pred(H1, H2), Mode, skolem(N, SnortRuleID), TimeRange)),	
        list_apply(ListOfOID, asset_mapping(N)).

summary(int(Pred(Host, Mode, skolem(N), TimeRange))):-
	setof(oid(SID,CID), Time^deriveIntPred(int(Pred(Host), Mode, Time), oid(SID,CID), SnortRuleID), ListOfOID),
	counter_next(sk, N),
	getTimeRangeForSummary(ListOfOID, TimeRange),
	assert(int(Pred(Host), Mode, skolem(N, SnortRuleID), TimeRange)),
        list_apply(ListOfOID, asset_mapping(N)).

getStartPoint(StartPoint):-
	db_query(snips,getstart,['select count(*) from snips_summary'], [StartPoint]).

counter_init_seed(Counter,S) :-
	CounterLiteral =.. [Counter, _],
	retractall(CounterLiteral),
	CounterInitLiteral =.. [Counter, S],
	assert(CounterInitLiteral).

counter_value(Counter,Value) :-
	CounterLiteral =.. [Counter,Value],
	CounterLiteral.

insertSummaryInDB:-
	iterate(int(Fact,Mode,skolem(SkolemId,SnortRuleId),range(StartTime,EndTime)),insertSummRowInDB(Fact,Mode,SkolemId,SnortRuleId,StartTime,EndTime)).
	
insertSummRowInDB(Fact,Mode,SkolemId,SnortRuleId,StartTime,EndTime):-
	Fact ^=.. [Pred,H],
	!,
        db_prepare(snips,ins,['insert into snips_summary  (pred,h1,h2,mode,skolem_id,snort_rule_id,start_time,end_time)  values(?,?,?,?,?,?,?,?)']),
        db_prepare_execute(ins, [Pred,H,'null',Mode,SkolemId,SnortRuleId,StartTime,EndTime], _P),
	db_statement_close(ins).

insertSummRowInDB(Fact,Mode,SkolemId,SnortRuleId,StartTime,EndTime):-
	Fact ^=.. [Pred,H1,H2],
        db_prepare(snips,ins,['insert into snips_summary  (pred,h1,h2,mode,skolem_id,snort_rule_id,start_time,end_time)  values(?,?,?,?,?,?,?,?)']),
        db_prepare_execute(ins, [Pred,H1,H2,Mode,SkolemId,SnortRuleId,StartTime,EndTime], _P),
	db_statement_close(ins).

insertMappingInDB:-
	iterate(mapping(SkolemID,SID,CID),insertMapRowInDB(SkolemID,SID,CID)).

insertMapRowInDB(SkolemID,SID,CID):-
        db_prepare(snips,ins,['insert into snips_skolem_map (skolem_id,sid,cid) values(?,?,?)']),
        db_prepare_execute(ins, [SkolemID,SID,CID], _P),
	db_statement_close(ins).

asset_mapping(N)(oid(SID,CID)) :-
	assert(mapping(N,SID,CID)).


%Map the observation to internal predicate using obsMap
deriveIntPred(int(Pred(H1, H2), Mode, Time), oid(SID,CID), SnortRuleID):-
	  obs(oid(SID,CID), Obs),
	  Obs=snort(SnortRuleID, _, _, _),
	  obsMap(_RuleId, obs(Obs), int(Fact), Mode),
	  Fact =.. [Pred, FH, TH, Time],
	  hostConf(FH, FHType),
	  hostConf(TH, THType),
	  summarizedIP(FH, FHType, H1),
	  summarizedIP(TH, THType, H2).


deriveIntPred(int(Pred(Host), Mode, Time), oid(SID,CID), SnortRuleID):-
	  obs(oid(SID,CID), Obs),
	  Obs=snort(SnortRuleID, _, _, _),
	  obsMap(_RuleId, obs(Obs), int(Fact), Mode),
	  Fact =.. [Pred, Host, Time].

/* DONE: Change list_first to find the minimum and maximum */

getTimeRangeForSummary(OidList, range(StartTime, EndTime)):-
	list_size(OidList, N),
	(N > 1 ->
	    min_timestamp(OidList, 1904834156, StartTime),
	    max_timestamp(OidList, 0 ,EndTime)
	;
	    list_first(OidList, FirstOid),
	    obs(FirstOid, snort(_, _, _, StartTime)),
	    obs(FirstOid, snort(_, _, _, EndTime))
	).
	      

summarizedIP(_H, external, external).
summarizedIP(H, internal, H).

min_timestamp([], MinTime, MinTime).

min_timestamp([Head|Rest], MinTime, Min):-
        obs(Head, snort(_, _, _,CurTime)),
        CurTime @< MinTime,
        min_timestamp(Rest, CurTime, Min).

min_timestamp([Head|Rest], MinTime, Min):-
        obs(Head, snort(_, _, _,CurTime)),
        CurTime @>= MinTime,
        min_timestamp(Rest, MinTime, Min).

max_timestamp([], MaxTime, MaxTime).

max_timestamp([Head|Rest], MaxTime, Max):-
        obs(Head, snort(_, _, _,CurTime)),
        CurTime @> MaxTime,
        max_timestamp(Rest, CurTime, Max).

max_timestamp([Head|Rest], MaxTime, Max):-
        obs(Head, snort(_, _, _,CurTime)),
        CurTime @=< MaxTime,
        max_timestamp(Rest, MaxTime, Max).

